from unittest import TestCase

from asm_6502 import Assembler


class TestAssembleCMP(TestCase):

    def setUp(self) -> None:
        self.assembler = Assembler()

    def test_cmp_immediate(self):
        code = "CMP #$10"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xC9, 0x10]),
        ], results)

    def test_cmp_zero_page(self):
        code = "CMP $10"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xC5, 0x10]),
        ], results)

    def test_cmp_zero_page_x(self):
        code = "CMP $10,X"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xD5, 0x10]),
        ], results)

    def test_cmp_absolute(self):
        code = "CMP $ABCD"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xCD, 0xCD, 0xAB]),
        ], results)

    def test_cmp_absolute_indexed(self):
        code = "CMP $ABCD,X\n" \
               "CMP $ABCD,Y"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xDD, 0xCD, 0xAB, 0xD9, 0xCD, 0xAB]),
        ], results)

    def test_cmp_indexed_indirect(self):
        code = "CMP ($10,X)"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xC1, 0x10]),
        ], results)

    def test_cmp_indirect_indexed(self):
        code = "CMP ($10),Y"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xD1, 0x10]),
        ], results)



from unittest import TestCase

from asm_6502 import Assembler


class TestAssembleADC(TestCase):

    def setUp(self) -> None:
        self.assembler = Assembler()

    def test_adc_immediate(self):
        code = "ADC #$10"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x69, 0x10]),
        ], results)

    def test_adc_zero_page(self):
        code = "ADC $10"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x65, 0x10]),
        ], results)

    def test_adc_zero_page_x(self):
        code = "ADC $10,X"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x75, 0x10]),
        ], results)

    def test_adc_absolute(self):
        code = "ADC $ABCD"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x6D, 0xCD, 0xAB]),
        ], results)

    def test_adc_absolute_indexed(self):
        code = "ADC $ABCD,X\n" \
               "ADC $ABCD,Y"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x7D, 0xCD, 0xAB, 0x79, 0xCD, 0xAB]),
        ], results)

    def test_adc_indexed_indirect(self):
        code = "ADC ($10,X)"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x61, 0x10]),
        ], results)

    def test_adc_indirect_indexed(self):
        code = "ADC ($10),Y"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x71, 0x10]),
        ], results)



from unittest import TestCase

from asm_6502 import Assembler


class TestAssembleISC(TestCase):

    def setUp(self) -> None:
        self.assembler = Assembler()

    def test_isc_zero_page(self):
        code = "ISC $00"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xE7, 0x00]),
        ], results)

    def test_isc_zero_page_x(self):
        code = "ISC $10,X"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xF7, 0x10]),
        ], results)

    def test_isc_absolute(self):
        code = "ISC $ABCD"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xEF, 0xCD, 0xAB]),
        ], results)

    def test_isc_absolute_indexed(self):
        code = "ISC $ABCD,X\n" \
               "ISC $ABCD,Y"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xFF, 0xCD, 0xAB, 0xFB, 0xCD, 0xAB]),
        ], results)

    def test_isc_indexed_indirect(self):
        code = "ISC ($40,X)"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xE3, 0x40]),
        ], results)

    def test_isc_indirect_indexed(self):
        code = "ISC ($40),Y"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xF3, 0x40]),
        ], results)



from unittest import TestCase

from asm_6502 import Assembler


class TestAssembleRRA(TestCase):

    def setUp(self) -> None:
        self.assembler = Assembler()

    def test_rra_zero_page(self):
        code = "RRA $00"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x67, 0x00]),
        ], results)

    def test_rra_zero_page_x(self):
        code = "RRA $10,X"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x77, 0x10]),
        ], results)

    def test_rra_absolute(self):
        code = "RRA $ABCD"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x6F, 0xCD, 0xAB]),
        ], results)

    def test_rra_absolute_indexed(self):
        code = "RRA $ABCD,X\n" \
               "RRA $ABCD,Y"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x7F, 0xCD, 0xAB, 0x7B, 0xCD, 0xAB]),
        ], results)

    def test_rra_indexed_indirect(self):
        code = "RRA ($40,X)"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x63, 0x40]),
        ], results)

    def test_rra_indirect_indexed(self):
        code = "RRA ($40),Y"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x73, 0x40]),
        ], results)



from unittest import TestCase

from asm_6502 import Assembler


class TestAssembleCPY(TestCase):

    def setUp(self) -> None:
        self.assembler = Assembler()

    def test_cpy_immediate(self):
        code = "CPY #$10"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xC0, 0x10]),
        ], results)

    def test_cpy_zero_page(self):
        code = "CPY $10"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xC4, 0x10]),
        ], results)

    def test_cpy_absolute(self):
        code = "CPY $ABCD"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xCC, 0xCD, 0xAB]),
        ], results)



from unittest import TestCase

from asm_6502 import Assembler


class TestAssembleROL(TestCase):

    def setUp(self) -> None:
        self.assembler = Assembler()

    def test_rol_accumulator(self):
        code = "ROL A"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x2A]),
        ], results)

    def test_rol_zero_page(self):
        code = "ROL $10"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x26, 0x10]),
        ], results)

    def test_rol_zero_page_x(self):
        code = "ROL $10,X"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x36, 0x10]),
        ], results)

    def test_rol_absolute(self):
        code = "ROL $ABCD"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x2E, 0xCD, 0xAB]),
        ], results)

    def test_rol_absolute_indexed_x(self):
        code = "ROL $ABCD,X"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x3E, 0xCD, 0xAB]),
        ], results)



from unittest import TestCase

from asm_6502 import Assembler


class TestAssembleCPX(TestCase):

    def setUp(self) -> None:
        self.assembler = Assembler()

    def test_cpx_immediate(self):
        code = "CPX #$10"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xE0, 0x10]),
        ], results)

    def test_cpx_zero_page(self):
        code = "CPX $10"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xE4, 0x10]),
        ], results)

    def test_cpx_absolute(self):
        code = "CPX $ABCD"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xEC, 0xCD, 0xAB]),
        ], results)



from unittest import TestCase

from asm_6502 import Assembler, AssembleError


class TestAssembleLAX(TestCase):

    def setUp(self) -> None:
        self.assembler = Assembler()

    def test_lax_immediate(self):
        code = "ORG $0080\n" \
               "LAX #10"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0080, [0xAB, 0x0A]),
        ], results)

    def test_lax_zero_page(self):
        code = "LAX $00"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xA7, 0x00]),
        ], results)

    def test_ldx_zero_page_x(self):
        code = "LAX $10,X"
        with self.assertRaises(AssembleError) as e:
            self.assembler.assemble(code)
        self.assertEqual("AssembleError: Can not use X as the index register in LAX at line 1",
                         str(e.exception))

    def test_lax_zero_page_y(self):
        code = "LAX $10,Y"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xB7, 0x10]),
        ], results)

    def test_lax_absolute(self):
        code = "LAX $ABCD"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xAF, 0xCD, 0xAB]),
        ], results)

    def test_lax_absolute_y(self):
        code = "LAX $ABCD,Y"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xBF, 0xCD, 0xAB]),
        ], results)

    def test_lax_indexed_indirect(self):
        code = "LAX ($40,X)"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xA3, 0x40]),
        ], results)

    def test_lax_indirect_indexed(self):
        code = "LAX ($40),Y"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xB3, 0x40]),
        ], results)



from unittest import TestCase

from asm_6502 import get_parser, Arithmetic, Integer


class TestParseArithmetic(TestCase):

    def setUp(self) -> None:
        self.parser = get_parser()

    def test_address(self):
        code = 'ORG  $0080'
        results = self.parser.parse(code)[0].addressing.address
        self.assertEqual(Integer(True, 128), results)

    def test_instant(self):
        code = 'TOLOWER LDY #$02'
        results = self.parser.parse(code)[0].addressing.address
        self.assertEqual(Integer(False, 2), results)

    def test_current(self):
        code = 'CMP *'
        results = self.parser.parse(code)[0].addressing.address
        self.assertEqual(Arithmetic(Arithmetic.CURRENT,), results)

    def test_add(self):
        code = "CMP #'Z'+1"
        results = self.parser.parse(code)[0].addressing.address
        self.assertEqual(Integer(False, 91), results)

    def test_sub(self):
        code = "ORA #%00100000-1"
        results = self.parser.parse(code)[0].addressing.address
        self.assertEqual(Integer(False, 31), results)

    def test_mul(self):
        code = 'CMP #2*3'
        results = self.parser.parse(code)[0].addressing.address
        self.assertEqual(Integer(False, 6), results)

        code = 'CMP #2*3+4*5'
        results = self.parser.parse(code)[0].addressing.address
        self.assertEqual(Integer(False, 26), results)

        code = 'CMP ***'
        results = self.parser.parse(code)[0].addressing.address
        self.assertEqual(Arithmetic(Arithmetic.MUL, (Arithmetic(Arithmetic.CURRENT), Arithmetic(Arithmetic.CURRENT))),
                         results)

    def test_div(self):
        code = 'CMP #24/3'
        results = self.parser.parse(code)[0].addressing.address
        self.assertEqual(Integer(False, 8), results)

        code = 'CMP */*'
        results = self.parser.parse(code)[0].addressing.address
        self.assertEqual(Arithmetic(Arithmetic.DIV, (Arithmetic(Arithmetic.CURRENT), Arithmetic(Arithmetic.CURRENT))),
                         results)

    def test_neg(self):
        code = 'CMP #-42'
        results = self.parser.parse(code)[0].addressing.address
        self.assertEqual(Integer(False, -42), results)

        code = 'CMP -**-*'
        results = self.parser.parse(code)[0].addressing.address
        self.assertEqual(Arithmetic(Arithmetic.MUL,
                                    (Arithmetic(Arithmetic.NEG, Arithmetic(Arithmetic.CURRENT)),
                                     Arithmetic(Arithmetic.NEG, Arithmetic(Arithmetic.CURRENT)))),
                         results)

    def test_parenthesis(self):
        code = 'CMP #2*[3+4*5]'
        results = self.parser.parse(code)[0].addressing.address
        self.assertEqual(Integer(False, 46), results)

    def test_low_and_high(self):
        code = 'LDA #LO $00AB+$CD00'
        results = self.parser.parse(code)[0].addressing.address
        self.assertEqual(Integer(False, 0xAB), results)

        code = 'LDA #HI $00AB+$CD00'
        results = self.parser.parse(code)[0].addressing.address
        self.assertEqual(Integer(False, 0xCD), results)

        code = 'LDA #LO *+1'
        results = self.parser.parse(code)[0].addressing.address
        self.assertEqual(Arithmetic(Arithmetic.LOW_BYTE,
                                    Arithmetic(Arithmetic.ADD, (Arithmetic(Arithmetic.CURRENT), Integer(False, 1)))),
                         results)

        code = 'LDA #HI *+1'
        results = self.parser.parse(code)[0].addressing.address
        self.assertEqual(Arithmetic(Arithmetic.HIGH_BYTE,
                                    Arithmetic(Arithmetic.ADD, (Arithmetic(Arithmetic.CURRENT), Integer(False, 1)))),
                         results)



from collections import namedtuple

import ply.lex as lex
import ply.yacc as yacc

__all__ = ['get_parser', 'ParseError', 'Integer', 'Addressing', 'Arithmetic', 'Instruction']


_PARSER = None
_LEXER = None


class Integer(namedtuple('Integer', ['is_word', 'value'])):

    __slots__ = ()

    def __add__(self, other):
        value = self.value + other.value
        return Integer(
            is_word=self.is_word or other.is_word or value > 0xFF,
            value=value,
        )

    def __sub__(self, other):
        return Integer(
            is_word=self.is_word or other.is_word,
            value=self.value - other.value,
        )

    def __mul__(self, other):
        value = self.value * other.value
        return Integer(
            is_word=self.is_word or other.is_word or value > 0xFF,
            value=value,
        )

    def __floordiv__(self, other):
        return Integer(
            is_word=self.is_word or other.is_word,
            value=self.value // other.value,
        )

    def __neg__(self):
        return Integer(is_word=self.is_word, value=-self.value)

    def low_byte(self):
        return Integer(is_word=False, value=self.value & 0xFF)

    def high_byte(self):
        return Integer(is_word=False, value=(self.value >> 8) & 0xFF)


class Addressing(namedtuple('Addressing', ['mode', 'address', 'register'], defaults=[None, None, None])):

    ACCUMULATOR = 'accumulator'
    IMMEDIATE = 'immediate'
    IMPLIED = 'implied'
    ADDRESS = 'address'
    ZERO_PAGE = 'zero page'
    ZERO_PAGE_X = 'zero page X'
    ZERO_PAGE_Y = 'zero page Y'
    ABSOLUTE = 'absolute'
    ABSOLUTE_X = 'absolute X'
    ABSOLUTE_Y = 'absolute Y'
    INDIRECT = 'indirect'
    INDEXED = 'indexed'
    INDEXED_INDIRECT = 'indexed indirect'
    INDIRECT_INDEXED = 'indirect indexed'

    LIST = 'list'


class Arithmetic(namedtuple('Arithmetic', ['mode', 'param'], defaults=[None, None])):

    CURRENT = 'current'
    LABEL = 'label'

    ADD = 'add'
    SUB = 'sub'
    MUL = 'mul'
    DIV = 'div'
    NEG = 'neg'

    LOW_BYTE = 'low_byte'
    HIGH_BYTE = 'high_byte'

    LIST = 'list'


class Instruction(namedtuple('Instruction', ['label', 'op', 'addressing', 'line_num'])):

    KEYWORDS = {
        'ADC', 'AND', 'ASL', 'BCC', 'BCS', 'BEQ', 'BIT', 'BMI', 'BNE', 'BPL', 'BRK', 'BVC', 'BVS', 'CLC',
        'CLD', 'CLI', 'CLV', 'CMP', 'CPX', 'CPY', 'DEC', 'DEX', 'DEY', 'EOR', 'INC', 'INX', 'INY', 'JMP',
        'JSR', 'LDA', 'LDX', 'LDY', 'LSR', 'NOP', 'ORA', 'PHA', 'PHP', 'PLA', 'PLP', 'ROL', 'ROR', 'RTI',
        'RTS', 'SBC', 'SEC', 'SED', 'SEI', 'STA', 'STX', 'STY', 'TAX', 'TAY', 'TSX', 'TXA', 'TXS', 'TYA',

        'ANC', 'ARR', 'ASR', 'DCP', 'ISC', 'JAM', 'LAS', 'LAX', 'RLA', 'RRA', 'SAX', 'SBX', 'SHA', 'SHS',
        'SHX', 'SHY', 'SLO', 'SRE', 'XAA'
    }

    PSEUDOS = {
        'ORG', '.ORG', '.BYTE', '.WORD', '.END'
    }


class ParseError(Exception):

    def __init__(self, info, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.info = info

    def __str__(self):
        return f'ParseError: {self.info}'

    def __repr__(self):
        return f'ParseError("{self.info}")'


def _get_column(p, index=None):
    column = 1
    if index is None:
        pos = p.lexpos
    else:
        pos = p.lexpos(index)
    while pos - column >= 0:
        if p.lexer.lexdata[pos - column] in {'\n', '\r'}:
            break
        column += 1
    return column


# Tokens
tokens = (
    'PSEUDO',
    'REGISTER',
    'KEYWORD',
    'LABEL',
    'BIT',
    'HEX',
    'BIN',
    'DEC',
    'CHAR',
    'CUR',
    'NEWLINE',
)

literals = ['+', '-', '/', '#', '(', ')', ',', '[', ']']


def t_PSEUDO(t):
    r"""\.[a-zA-Z_][a-zA-Z0-9_]*"""
    t.type = 'KEYWORD'
    return t


def t_LABEL(t):
    r"""[a-zA-Z_][a-zA-Z0-9_]*"""
    if t.value in {'A', 'X', 'Y'}:
        t.type = 'REGISTER'
    elif t.value in Instruction.KEYWORDS or t.value in Instruction.PSEUDOS:
        t.type = 'KEYWORD'
    return t


t_BIT = r'\#LO|\#HI'


def t_HEX(t):
    r"""\$[0-9a-fA-F]+"""
    t.value = Integer(
        is_word=len(t.value) > 2 + 1,
        value=int(t.value[1:], 16),
    )
    return t


def t_BIN(t):
    r"""%[01]+"""
    t.value = Integer(
        is_word=len(t.value) > 8 + 1,
        value=int(t.value[1:], 2),
    )
    return t


def t_DEC(t):
    r"""[0-9]+"""
    value = int(t.value, 10)
    t.value = Integer(
        is_word=value > 0xFF,
        value=value,
    )
    return t


def t_CHAR(t):
    r"""\'[^\'\n\r]\'"""
    t.value = Integer(is_word=False, value=ord(t.value[1]))
    return t


t_CUR = r'\*'

t_ignore = " \t"
t_ignore_COMMENT = r';.*'


def t_NEWLINE(t):
    r"""[\n\r]+"""
    t.lexer.lineno += t.value.count("\n")
    return t


def t_error(t):
    raise ParseError(f"Illegal character '{t.value[0]}' found at line {t.lineno}, column {_get_column(t)}")


# Syntax
precedence = (
    ('left', '+', '-'),
    ('left', 'CUR', '/'),
    ('right', 'UMINUS'),
)


def p_stat_with_label(p):
    """stat : LABEL KEYWORD stat_val"""
    p[0] = [Instruction(label=p[1], op=p[2], addressing=p[3], line_num=p.lineno(1))]
    return p


def p_stat_without_label(p):
    """stat : KEYWORD stat_val"""
    p[0] = [Instruction(label=None, op=p[1], addressing=p[2], line_num=p.lineno(1))]
    return p


def p_stat_repeat(p):
    """stat : stat NEWLINE stat"""
    p[0] = p[1] + p[3]
    return p


def p_stat_empty(p):
    """stat :"""
    p[0] = []
    return p


def p_stat_val_accumulator(p):
    """stat_val : REGISTER"""
    if p[1] == 'A':
        p[0] = Addressing(Addressing.ACCUMULATOR)
    else:
        raise ParseError(f"Register {p[1]} can not be used as an address at "
                         f"{p.lineno(1)}, column {_get_column(p, index=1)}")
    return p


def p_stat_val_direct(p):
    """stat_val : arithmetic"""
    p[0] = Addressing(Addressing.ADDRESS, address=p[1])
    return p


def p_stat_val_empty(p):
    """stat_val :"""
    p[0] = Addressing(Addressing.IMPLIED,)
    return p


def p_stat_val_indirect(p):
    """stat_val : '(' arithmetic ')'"""
    p[0] = Addressing(Addressing.INDIRECT, address=p[2])
    return p


def p_stat_val_indexed(p):
    """stat_val : arithmetic ',' REGISTER"""
    if p[3][0] not in {'X', 'Y'}:
        raise ParseError(f"Only registers X and Y can be used for indexing, found {p[3][0]} at "
                         f"{p.lineno(3)}, column {_get_column(p, index=3)}")
    p[0] = Addressing(Addressing.INDEXED, address=p[1], register=p[3])
    return p


def p_stat_val_indexed_indirect(p):
    """stat_val : '(' arithmetic ',' REGISTER ')'"""
    if p[4][0] != 'X':
        raise ParseError(f"Only register X can be used for indexed indirect addressing, found {p[4][0]} at "
                         f"{p.lineno(4)}, column {_get_column(p, index=4)}")
    p[0] = Addressing(Addressing.INDEXED_INDIRECT, address=p[2], register=p[4])
    return p


def p_stat_val_indirect_indexed(p):
    """stat_val : '(' arithmetic ')' ',' REGISTER"""
    if p[5][0] != 'Y':
        raise ParseError(f"Only register Y can be used for indexed indirect addressing, found {p[5][0]} at "
                         f"{p.lineno(5)}, column {_get_column(p, index=5)}")
    p[0] = Addressing(Addressing.INDIRECT_INDEXED, address=p[2], register=p[5])
    return p


def p_stat_val_immediate_bit(p):
    """stat_val : BIT arithmetic"""
    if p[1] == '#LO':
        if isinstance(p[2], Integer):
            p[0] = Addressing(Addressing.IMMEDIATE, address=p[2].low_byte())
        else:
            p[0] = Addressing(Addressing.IMMEDIATE, address=Arithmetic(Arithmetic.LOW_BYTE, p[2]))
    else:
        if isinstance(p[2], Integer):
            p[0] = Addressing(Addressing.IMMEDIATE, address=p[2].high_byte())
        else:
            p[0] = Addressing(Addressing.IMMEDIATE, address=Arithmetic(Arithmetic.HIGH_BYTE, p[2]))
    return p


def p_stat_val_immediate(p):
    """stat_val : '#' arithmetic"""
    p[0] = Addressing(Addressing.IMMEDIATE, address=p[2])
    return p


def p_stat_val_list(p):
    """stat_val : arithmetic_list"""
    p[0] = Addressing(Addressing.LIST, address=p[1])
    return p


def p_arithmetic_list(p):
    """arithmetic_list : arithmetic ',' arithmetic_list
                       | arithmetic"""
    if len(p) == 2:
        p[0] = Arithmetic(Arithmetic.LIST, [p[1]])
    else:
        p[0] = Arithmetic(Arithmetic.LIST, [p[1]] + p[3].param)
    return p


def p_arithmetic_uminus(p):
    """arithmetic : '-' arithmetic %prec UMINUS"""
    if isinstance(p[2], Integer):
        p[0] = -p[2]
    else:
        p[0] = Arithmetic(Arithmetic.NEG, p[2])
    return p


def p_arithmetic_direct(p):
    """arithmetic : integer"""
    p[0] = p[1]
    return p


def p_arithmetic_label(p):
    """arithmetic : LABEL"""
    p[0] = Arithmetic(Arithmetic.LABEL, p[1])
    return p


def p_arithmetic_cur(p):
    """arithmetic : CUR"""
    p[0] = Arithmetic(Arithmetic.CURRENT)
    return p


def p_arithmetic_paren(p):
    """arithmetic : '[' arithmetic ']'"""
    p[0] = p[2]
    return p


def p_arithmetic_binary_op(p):
    """arithmetic : arithmetic '+' arithmetic
                  | arithmetic '-' arithmetic
                  | arithmetic CUR arithmetic
                  | arithmetic '/' arithmetic
    """
    if isinstance(p[1], Integer) and isinstance(p[3], Integer):
        if p[2] == '+':
            p[0] = p[1] + p[3]
        elif p[2] == '-':
            p[0] = p[1] - p[3]
        elif p[2] == '/':
            p[0] = p[1] // p[3]
        else:
            p[0] = p[1] * p[3]
    else:
        if p[2] == '+':
            p[0] = Arithmetic(Arithmetic.ADD, (p[1], p[3]))
        elif p[2] == '-':
            p[0] = Arithmetic(Arithmetic.SUB, (p[1], p[3]))
        elif p[2] == '/':
            p[0] = Arithmetic(Arithmetic.DIV, (p[1], p[3]))
        else:
            p[0] = Arithmetic(Arithmetic.MUL, (p[1], p[3]))
    return p


def p_integer(p):
    """integer : DEC
              | HEX
              | BIN
              | CHAR
    """
    p[0] = p[1]
    return p


def p_error(p):
    if p:
        raise ParseError(f"Syntax error at line {p.lineno}, column {_get_column(p)}: {repr(p.value)}")
    else:
        raise ParseError(f"Syntax error at EOF")


def get_parser(debug=False):
    global _PARSER, _LEXER
    if _PARSER is None:
        _LEXER = lex.lex(debug=debug)
        _PARSER = yacc.yacc(debug=debug)
    _LEXER.lineno = 1
    return _PARSER



from unittest import TestCase

from asm_6502 import Assembler, AssembleError


class TestAssembleNOP(TestCase):

    def setUp(self) -> None:
        self.assembler = Assembler()

    def test_nop_implied(self):
        code = "NOP"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xEA]),
        ], results)

    def test_nop_immediate_undocumented(self):
        code = "NOP #$10"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x80, 0x10]),
        ], results)

    def test_nop_zero_page_undocumented(self):
        code = "NOP $10"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x04, 0x10]),
        ], results)

    def test_nop_zero_page_x_undocumented(self):
        code = "NOP $10,X"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x14, 0x10]),
        ], results)

    def test_nop_absolute_undocumented(self):
        code = "NOP $ABCD"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x0C, 0xCD, 0xAB]),
        ], results)

    def test_nop_absolute_x_undocumented(self):
        code = "NOP $ABCD,X"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x1C, 0xCD, 0xAB]),
        ], results)

    def test_nop_absolute_y_undocumented(self):
        code = "NOP $ABCD,Y"
        with self.assertRaises(AssembleError) as e:
            self.assembler.assemble(code)
        self.assertEqual("AssembleError: Can not use Y as the index register in NOP at line 1",
                         str(e.exception))



from unittest import TestCase

from asm_6502 import Assembler


class TestAssembleTAY(TestCase):

    def setUp(self) -> None:
        self.assembler = Assembler()

    def test_tay_implied(self):
        code = "TAY"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xA8]),
        ], results)



from unittest import TestCase

from asm_6502 import Assembler, AssembleError


class TestAssembleWORD(TestCase):

    def setUp(self) -> None:
        self.assembler = Assembler()

    def test_word(self):
        code = ".ORG $1000\n" \
               ".WORD $ABCD"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x1000, [0xCD, 0xAB]),
        ], results)

    def test_word_with_arithmetic(self):
        code = ".ORG $1000\n" \
               ".WORD *+3"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x1000, [0x03, 0x10]),
        ], results)

    def test_words(self):
        code = ".ORG $1000\n" \
               ".WORD $ABCD, $EF42"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x1000, [0xCD, 0xAB, 0x42, 0xEF]),
        ], results)



from unittest import TestCase

from asm_6502 import get_parser


class TestParseWikiSample(TestCase):

    def test_wiki_sample(self):
        sample = """
; TOLOWER:
;
;   Convert a null-terminated character string to all lower case.
;   Maximum string length is 255 characters, plus the null term-
;   inator.
;
; Parameters:
;
;   SRC - Source string address
;   DST - Destination string address
;
        ORG  $0080
;
SRC     .WORD $0400     ;source string pointer
DST     .WORD $0500     ;destination string pointer
;
        ORG $0600       ;execution start address
;
TOLOWER LDY #$00        ;starting index
;
LOOP    LDA (SRC),Y     ;get from source string
        BEQ DONE        ;end of string
;
        CMP #'A'        ;if lower than UC alphabet...
        BCC SKIP        ;copy unchanged
;
        CMP #'Z'+1      ;if greater than UC alphabet...
        BCS SKIP        ;copy unchanged
;
        ORA #%00100000  ;convert to lower case
;
SKIP    STA (DST),Y     ;store to destination string
        INY             ;bump index
        BNE LOOP        ;next character
;
; NOTE: If Y wraps the destination string will be left in an undefined
;  state.  We set carry to indicate this to the calling function.
;
        SEC             ;report string too long error &...
        RTS             ;return to caller
;
DONE    STA (DST),Y     ;terminate destination string
        CLC             ;report conversion completed &...
        RTS             ;return to caller
;
        .END
"""
        parser = get_parser(debug=False)
        parser.parse(sample)



from unittest import TestCase

from asm_6502 import Assembler


class TestAssembleSLO(TestCase):

    def setUp(self) -> None:
        self.assembler = Assembler()

    def test_slo_zero_page(self):
        code = "SLO $00"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x07, 0x00]),
        ], results)

    def test_slo_zero_page_x(self):
        code = "SLO $10,X"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x17, 0x10]),
        ], results)

    def test_slo_absolute(self):
        code = "SLO $ABCD"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x0F, 0xCD, 0xAB]),
        ], results)

    def test_slo_absolute_indexed(self):
        code = "SLO $ABCD,X\n" \
               "SLO $ABCD,Y"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x1F, 0xCD, 0xAB, 0x1B, 0xCD, 0xAB]),
        ], results)

    def test_slo_indexed_indirect(self):
        code = "SLO ($40,X)"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x03, 0x40]),
        ], results)

    def test_slo_indirect_indexed(self):
        code = "SLO ($40),Y"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x13, 0x40]),
        ], results)



from unittest import TestCase

from asm_6502 import Assembler


class TestAssembleAND(TestCase):

    def setUp(self) -> None:
        self.assembler = Assembler()

    def test_and_immediate(self):
        code = "AND #$10"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x29, 0x10]),
        ], results)

    def test_and_zero_page(self):
        code = "AND $10"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x25, 0x10]),
        ], results)

    def test_and_zero_page_x(self):
        code = "AND $10,X"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x35, 0x10]),
        ], results)

    def test_and_absolute(self):
        code = "AND $ABCD"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x2D, 0xCD, 0xAB]),
        ], results)

    def test_and_absolute_indexed(self):
        code = "AND $ABCD,X\n" \
               "AND $ABCD,Y"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x3D, 0xCD, 0xAB, 0x39, 0xCD, 0xAB]),
        ], results)

    def test_and_indexed_indirect(self):
        code = "AND ($10,X)"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x21, 0x10]),
        ], results)

    def test_and_indirect_indexed(self):
        code = "AND ($10),Y"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x31, 0x10]),
        ], results)



from unittest import TestCase

from asm_6502 import Assembler


class TestAssembleEND(TestCase):

    def setUp(self) -> None:
        self.assembler = Assembler()

    def test_end(self):
        code = ".ORG $1000\n" \
               ".END"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x1000, [0x4C, 0x00, 0x10]),
        ], results)



from unittest import TestCase

from asm_6502 import Assembler


class TestAssembleSRE(TestCase):

    def setUp(self) -> None:
        self.assembler = Assembler()

    def test_sre_zero_page(self):
        code = "SRE $00"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x47, 0x00]),
        ], results)

    def test_sre_zero_page_x(self):
        code = "SRE $10,X"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x57, 0x10]),
        ], results)

    def test_sre_absolute(self):
        code = "SRE $ABCD"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x4F, 0xCD, 0xAB]),
        ], results)

    def test_sre_absolute_indexed(self):
        code = "SRE $ABCD,X\n" \
               "SRE $ABCD,Y"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x5F, 0xCD, 0xAB, 0x5B, 0xCD, 0xAB]),
        ], results)

    def test_sre_indexed_indirect(self):
        code = "SRE ($40,X)"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x43, 0x40]),
        ], results)

    def test_sre_indirect_indexed(self):
        code = "SRE ($40),Y"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x53, 0x40]),
        ], results)



from unittest import TestCase

from asm_6502 import Assembler, AssembleError


class TestAssembleBCC(TestCase):

    def setUp(self) -> None:
        self.assembler = Assembler()

    def test_bcc_relative(self):
        code = "BCC *+$10"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x90, 0x0E]),
        ], results)

    def test_bcc_relative_neg(self):
        code = "ORG $0800\n" \
               "BCC *-$7E"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0800, [0x90, 0x80]),
        ], results)

    def test_bcc_relative_neg_invalid(self):
        code = "ORG $0800\n" \
               "BCC *-$7F"
        with self.assertRaises(AssembleError) as e:
            self.assembler.assemble(code)
        self.assertEqual("AssembleError: The offset -0x81 is out of range for relative addressing at line 2",
                         str(e.exception))

    def test_bcc_relative_pos(self):
        code = "ORG $0800\n" \
               "BCC *+$81"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0800, [0x90, 0x7F]),
        ], results)

    def test_bcc_relative_pos_invalid(self):
        code = "ORG $0800\n" \
               "BCC *+$82"
        with self.assertRaises(AssembleError) as e:
            self.assembler.assemble(code)
        self.assertEqual("AssembleError: The offset 0x80 is out of range for relative addressing at line 2",
                         str(e.exception))



from unittest import TestCase

from asm_6502 import Assembler


class TestAssembleJSR(TestCase):

    def setUp(self) -> None:
        self.assembler = Assembler()

    def test_jsr_absolute(self):
        code = "START ORG $0080\n" \
               "START ORG $0800\n" \
               "      JSR START"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0800, [0x20, 0x00, 0x08]),
        ], results)



from unittest import TestCase

from asm_6502 import Assembler


class TestAssembleROR(TestCase):

    def setUp(self) -> None:
        self.assembler = Assembler()

    def test_ror_accumulator(self):
        code = "ROR A"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x6A]),
        ], results)

    def test_ror_zero_page(self):
        code = "ROR $10"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x66, 0x10]),
        ], results)

    def test_ror_zero_page_x(self):
        code = "ROR $10,X"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x76, 0x10]),
        ], results)

    def test_ror_absolute(self):
        code = "ROR $ABCD"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x6E, 0xCD, 0xAB]),
        ], results)

    def test_ror_absolute_indexed_x(self):
        code = "ROR $ABCD,X"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x7E, 0xCD, 0xAB]),
        ], results)



from unittest import TestCase

from asm_6502 import Assembler, AssembleError


class TestAssembleSHY(TestCase):

    def setUp(self) -> None:
        self.assembler = Assembler()

    def test_shy_absolute_x(self):
        code = "SHY $ABCD,X"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x9C, 0xCD, 0xAB]),
        ], results)

    def test_shy_absolute_y(self):
        code = "SHY $1000,Y"
        with self.assertRaises(AssembleError) as e:
            self.assembler.assemble(code)
        self.assertEqual("AssembleError: Can not use Y as the index register in SHY at line 1",
                         str(e.exception))



from unittest import TestCase

from asm_6502 import Assembler, AssembleError


class TestAssembleBYTE(TestCase):

    def setUp(self) -> None:
        self.assembler = Assembler()

    def test_byte(self):
        code = ".ORG $1000\n" \
               ".BYTE $AB"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x1000, [0xAB]),
        ], results)

    def test_byte_too_large(self):
        code = ".ORG $1000\n" \
               ".BYTE $ABCD"
        with self.assertRaises(AssembleError) as e:
            self.assembler.assemble(code)
        self.assertEqual("AssembleError: 0xabcd can not fit in a byte at line 2", str(e.exception))

    def test_bytes(self):
        code = ".ORG $1000\n" \
               ".BYTE $AB,$CD+1"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x1000, [0xAB, 0xCD + 1]),
        ], results)

        code = ".ORG $1000\n" \
               ".BYTE $AB,$CD+1,$EF"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x1000, [0xAB, 0xCD + 1, 0xEF]),
        ], results)

    def test_bytes_too_large(self):
        code = ".ORG $1000\n" \
               ".BYTE $AB,$CD+$EF"
        with self.assertRaises(AssembleError) as e:
            self.assembler.assemble(code)
        self.assertEqual("AssembleError: 0x1bc can not fit in a byte at line 2", str(e.exception))



from unittest import TestCase

from asm_6502 import Assembler, AssembleError


class TestAssembleJMP(TestCase):

    def setUp(self) -> None:
        self.assembler = Assembler()

    def test_jmp_absolute(self):
        code = "ORG $0080\n" \
               "JMP $0080"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0080, [0x4C, 0x80, 0x00]),
        ], results)

        code = "START ORG $0080\n" \
               "      JMP START"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0080, [0x4C, 0x80, 0x00]),
        ], results)

    def test_jmp_indirect(self):
        code = "START ORG $0080\n" \
               "      JMP (START)\n" \
               "      ORG $FFFF"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0080, [0x6C, 0x80, 0x00]),
        ], results)

    def test_jmp_error_addressing(self):
        code = "JMP ($0080),Y"
        with self.assertRaises(AssembleError) as e:
            self.assembler.assemble(code)
        self.assertEqual("AssembleError: Indirect indexed addressing is not allowed for `JMP` at line 1",
                         str(e.exception))

    def test_jmp_unknown_label(self):
        code = "JMP START"
        with self.assertRaises(AssembleError) as e:
            self.assembler.assemble(code)
        self.assertEqual("AssembleError(\"Can not resolve label 'START' at line 1\")",
                         repr(e.exception))

    def test_jmp_out_of_memory(self):
        code = "START ORG $FFFF\n" \
               "      JMP (START)"
        with self.assertRaises(AssembleError) as e:
            self.assembler.assemble(code)
        self.assertEqual("AssembleError: The assembled code will exceed the max memory 0x10000 at line 2",
                         str(e.exception))

    def test_jmp_with_entry(self):
        code = "ORG $0080\n" \
               "JMP $abcd"
        results = self.assembler.assemble(code)
        self.assertEqual([
            (0x0080, [0x4C, 0xcd, 0xab]),
            (0xFFFC, [0x80, 0x00]),
        ], results)

        code = "START ORG $0080\n" \
               "NEXT  JMP $abcd"
        results = self.assembler.assemble(code)
        self.assertEqual([
            (0x0080, [0x4C, 0xcd, 0xab]),
            (0xFFFC, [0x80, 0x00]),
        ], results)

    def test_jmp_dead_loop(self):
        code = "ORG $0080\n" \
               "JMP *"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0080, [0x4C, 0x80, 0x00]),
        ], results)

    def test_jmp_add(self):
        code = "ORG $0080\n" \
               "JMP *+3"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0080, [0x4C, 0x83, 0x00]),
        ], results)

    def test_jmp_sub(self):
        code = "ORG $0080\n" \
               "JMP *-3"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0080, [0x4C, 0x7D, 0x00]),
        ], results)

    def test_jmp_mul(self):
        code = "ORG $0080\n" \
               "JMP ***"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0080, [0x4C, 0x00, 0x40]),
        ], results)

    def test_jmp_div(self):
        code = "START ORG $0080\n" \
               "      JMP */START"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0080, [0x4C, 0x01, 0x00]),
        ], results)



from unittest import TestCase

from asm_6502 import Assembler, AssembleError


class TestAssembleLDA(TestCase):

    def setUp(self) -> None:
        self.assembler = Assembler()

    def test_lda_immediate(self):
        code = "ORG $0080\n" \
               "LDA #10"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0080, [0xA9, 0x0A]),
        ], results)

    def test_lda_immediate_too_large(self):
        code = "ORG $0080\n" \
               "LDA #$ABCD"
        with self.assertRaises(AssembleError) as e:
            self.assembler.assemble(code)
        self.assertEqual("AssembleError: The value 0xabcd is too large for the addressing at line 2",
                         str(e.exception))

    def test_lda_immediate_low(self):
        code = "START ORG $ABCD\n" \
               "      LDA #LO START"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0xABCD, [0xA9, 0xCD]),
        ], results)

        code = "LDA #LO $ABCD"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xA9, 0xCD]),
        ], results)

    def test_lda_immediate_high(self):
        code = "START ORG $ABCD\n" \
               "      LDA #HI START"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0xABCD, [0xA9, 0xAB]),
        ], results)

    def test_lda_zero_page(self):
        code = "LDA $00\n" \
               "LDA $00"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xA5, 0x00, 0xA5, 0x00]),
        ], results)

        code = "START LDA START"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xAD, 0x00, 0x00]),
        ], results)

        code = "ORG $ABCD\n" \
               "LDA *-*"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0xABCD, [0xAD, 0x00, 0x00]),
        ], results)

        code = "ORG $ABCD\n" \
               "LDA -*--*"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0xABCD, [0xAD, 0x00, 0x00]),
        ], results)

        code = "LDA *+255"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xAD, 0xFF, 0x00]),
        ], results)

    def test_lda_zero_page_x(self):
        code = "LDA $10,X\n" \
               "LDA $10,X"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xB5, 0x10, 0xB5, 0x10]),
        ], results)

    def test_lda_absolute(self):
        code = "LDA $ABCD\n" \
               "LDA $ABCD"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xAD, 0xCD, 0xAB, 0xAD, 0xCD, 0xAB]),
        ], results)

        code = "LDA $00CD\n" \
               "LDA $00CD"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xAD, 0xCD, 0x00, 0xAD, 0xCD, 0x00]),
        ], results)

        code = "LDA $FF+10-10"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xAD, 0xFF, 0x00]),
        ], results)

    def test_lda_absolute_indexed(self):
        code = "LDA $ABCD,X\n" \
               "LDA $ABCD,Y"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xBD, 0xCD, 0xAB, 0xB9, 0xCD, 0xAB]),
        ], results)

        code = "LDA $10,X\n" \
               "LDA $10,Y"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xB5, 0x10, 0xB9, 0x10, 0x00]),
        ], results)

        code = "LDA $0010,X\n" \
               "LDA $0010,Y"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xBD, 0x10, 0x00, 0xB9, 0x10, 0x00]),
        ], results)

    def test_lda_indexed_indirect(self):
        code = "LDA ($40,X)"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xA1, 0x40]),
        ], results)

    def test_lda_indirect_indexed(self):
        code = "LDA ($40),Y"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xB1, 0x40]),
        ], results)



from unittest import TestCase

from asm_6502 import Assembler


class TestAssembleEOR(TestCase):

    def setUp(self) -> None:
        self.assembler = Assembler()

    def test_eor_immediate(self):
        code = "EOR #$10"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x49, 0x10]),
        ], results)

    def test_eor_zero_page(self):
        code = "EOR $10"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x45, 0x10]),
        ], results)

    def test_eor_zero_page_x(self):
        code = "EOR $10,X"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x55, 0x10]),
        ], results)

    def test_eor_absolute(self):
        code = "EOR $ABCD"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x4D, 0xCD, 0xAB]),
        ], results)

    def test_eor_absolute_indexed(self):
        code = "EOR $ABCD,X\n" \
               "EOR $ABCD,Y"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x5D, 0xCD, 0xAB, 0x59, 0xCD, 0xAB]),
        ], results)

    def test_eor_indexed_indirect(self):
        code = "EOR ($10,X)"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x41, 0x10]),
        ], results)

    def test_eor_indirect_indexed(self):
        code = "EOR ($10),Y"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x51, 0x10]),
        ], results)



from unittest import TestCase

from asm_6502 import Assembler


class TestAssembleRLA(TestCase):

    def setUp(self) -> None:
        self.assembler = Assembler()

    def test_rla_zero_page(self):
        code = "RLA $00"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x27, 0x00]),
        ], results)

    def test_rla_zero_page_x(self):
        code = "RLA $10,X"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x37, 0x10]),
        ], results)

    def test_rla_absolute(self):
        code = "RLA $ABCD"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x2F, 0xCD, 0xAB]),
        ], results)

    def test_rla_absolute_indexed(self):
        code = "RLA $ABCD,X\n" \
               "RLA $ABCD,Y"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x3F, 0xCD, 0xAB, 0x3B, 0xCD, 0xAB]),
        ], results)

    def test_rla_indexed_indirect(self):
        code = "RLA ($40,X)"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x23, 0x40]),
        ], results)

    def test_rla_indirect_indexed(self):
        code = "RLA ($40),Y"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x33, 0x40]),
        ], results)



from unittest import TestCase

from asm_6502 import Assembler, AssembleError


class TestAssembleSHA(TestCase):

    def setUp(self) -> None:
        self.assembler = Assembler()

    def test_sha_absolute_x(self):
        code = "SHA $1000,X"
        with self.assertRaises(AssembleError) as e:
            self.assembler.assemble(code)
        self.assertEqual("AssembleError: Can not use X as the index register in SHA at line 1",
                         str(e.exception))

    def test_sha_absolute_y(self):
        code = "SHA $ABCD,Y"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x9F, 0xCD, 0xAB]),
        ], results)

    def test_sha_indirect_indexed(self):
        code = "SHA ($AB),Y"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x93, 0xAB]),
        ], results)



from unittest import TestCase

from asm_6502 import Assembler, AssembleError


class TestAssembleSTY(TestCase):

    def setUp(self) -> None:
        self.assembler = Assembler()

    def test_sty_zero_page(self):
        code = "STY $10"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x84, 0x10]),
        ], results)

    def test_sty_zero_page_x(self):
        code = "STY $10,X"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x94, 0x10]),
        ], results)

    def test_sty_zero_page_y(self):
        code = "STY $10,Y"
        with self.assertRaises(AssembleError) as e:
            self.assembler.assemble(code)
        self.assertEqual("AssembleError: Can not use Y as the index register in STY at line 1",
                         str(e.exception))

    def test_sty_absolute(self):
        code = "STY $ABCD"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x8C, 0xCD, 0xAB]),
        ], results)

    def test_sty_absolute_indexed(self):
        code = "STY $ABCD,X"
        with self.assertRaises(AssembleError) as e:
            self.assembler.assemble(code)
        self.assertEqual("AssembleError: Absolute indexed addressing is not allowed for STY at line 1",
                         str(e.exception))



from unittest import TestCase

from asm_6502 import Assembler


class TestAssembleSBC(TestCase):

    def setUp(self) -> None:
        self.assembler = Assembler()

    def test_sbc_immediate(self):
        code = "SBC #$10"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xE9, 0x10]),
        ], results)

    def test_sbc_zero_page(self):
        code = "SBC $10"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xE5, 0x10]),
        ], results)

    def test_sbc_zero_page_x(self):
        code = "SBC $10,X"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xF5, 0x10]),
        ], results)

    def test_sbc_absolute(self):
        code = "SBC $ABCD"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xED, 0xCD, 0xAB]),
        ], results)

    def test_sbc_absolute_indexed(self):
        code = "SBC $ABCD,X\n" \
               "SBC $ABCD,Y"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xFD, 0xCD, 0xAB, 0xF9, 0xCD, 0xAB]),
        ], results)

    def test_sbc_indexed_indirect(self):
        code = "SBC ($10,X)"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xE1, 0x10]),
        ], results)

    def test_sbc_indirect_indexed(self):
        code = "SBC ($10),Y"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xF1, 0x10]),
        ], results)



from unittest import TestCase

from asm_6502 import Assembler, AssembleError


class TestAssembleSTA(TestCase):

    def setUp(self) -> None:
        self.assembler = Assembler()

    def test_sta_error_immediate(self):
        code = "STA #$00"
        with self.assertRaises(AssembleError) as e:
            self.assembler.assemble(code)
        self.assertEqual("AssembleError: Immediate addressing is not allowed for `STA` at line 1",
                         str(e.exception))

    def test_sta_zero_page(self):
        code = "STA $00"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x85, 0x00]),
        ], results)

    def test_sta_zero_page_x(self):
        code = "STA $10,X"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x95, 0x10]),
        ], results)

    def test_sta_absolute(self):
        code = "STA $ABCD"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x8D, 0xCD, 0xAB]),
        ], results)

    def test_sta_absolute_indexed(self):
        code = "STA $ABCD,X\n" \
               "STA $ABCD,Y"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x9D, 0xCD, 0xAB, 0x99, 0xCD, 0xAB]),
        ], results)

    def test_sta_indexed_indirect(self):
        code = "STA ($40,X)"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x81, 0x40]),
        ], results)

    def test_sta_indirect_indexed(self):
        code = "STA ($40),Y"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x91, 0x40]),
        ], results)



from unittest import TestCase

from asm_6502 import Assembler


class TestAssembleDCP(TestCase):

    def setUp(self) -> None:
        self.assembler = Assembler()

    def test_dcp_zero_page(self):
        code = "DCP $00"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xC7, 0x00]),
        ], results)

    def test_dcp_zero_page_x(self):
        code = "DCP $10,X"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xD7, 0x10]),
        ], results)

    def test_dcp_absolute(self):
        code = "DCP $ABCD"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xCF, 0xCD, 0xAB]),
        ], results)

    def test_dcp_absolute_indexed(self):
        code = "DCP $ABCD,X\n" \
               "DCP $ABCD,Y"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xDF, 0xCD, 0xAB, 0xDB, 0xCD, 0xAB]),
        ], results)

    def test_dcp_indexed_indirect(self):
        code = "DCP ($40,X)"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xC3, 0x40]),
        ], results)

    def test_dcp_indirect_indexed(self):
        code = "DCP ($40),Y"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0xD3, 0x40]),
        ], results)



from unittest import TestCase

from asm_6502 import Assembler


class TestAssembleORA(TestCase):

    def setUp(self) -> None:
        self.assembler = Assembler()

    def test_ora_immediate(self):
        code = "ORA #$10"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x09, 0x10]),
        ], results)

    def test_ora_zero_page(self):
        code = "ORA $10"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x05, 0x10]),
        ], results)

    def test_ora_zero_page_x(self):
        code = "ORA $10,X"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x15, 0x10]),
        ], results)

    def test_ora_absolute(self):
        code = "ORA $ABCD"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x0D, 0xCD, 0xAB]),
        ], results)

    def test_ora_absolute_indexed(self):
        code = "ORA $ABCD,X\n" \
               "ORA $ABCD,Y"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x1D, 0xCD, 0xAB, 0x19, 0xCD, 0xAB]),
        ], results)

    def test_ora_indexed_indirect(self):
        code = "ORA ($10,X)"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x01, 0x10]),
        ], results)

    def test_ora_indirect_indexed(self):
        code = "ORA ($10),Y"
        results = self.assembler.assemble(code, add_entry=False)
        self.assertEqual([
            (0x0000, [0x11, 0x10]),
        ], results)



from unittest import TestCase

from asm_6502 import get_parser, Addressing, Integer


class TestParseAddressing(TestCase):

    def setUp(self) -> None:
        self.parser = get_parser()

    def test_addressing_accumulator(self):
        code = 'LSR A'
        results = self.parser.parse(code)[0].addressing
        self.assertEqual(Addressing(Addressing.ACCUMULATOR,), results)

    def test_addressing_immediate(self):
        code = 'ORA #$B2'
        results = self.parser.parse(code)[0].addressing
        self.assertEqual(Addressing(Addressing.IMMEDIATE, address=Integer(False, 178)), results)

    def test_addressing_implied(self):
        code = 'CLC'
        results = self.parser.parse(code)[0].addressing
        self.assertEqual(Addressing(Addressing.IMPLIED,), results)

    def test_addressing_address(self):
        code = 'JMP $4032'
        results = self.parser.parse(code)[0].addressing
        self.assertEqual(Addressing(Addressing.ADDRESS, address=Integer(True, 16434)), results)

        code = 'LDA $35'
        results = self.parser.parse(code)[0].addressing
        self.assertEqual(Addressing(Addressing.ADDRESS, address=Integer(False, 53)), results)

    def test_addressing_indirect(self):
        code = 'JMP  ($1000)'
        results = self.parser.parse(code)[0].addressing
        self.assertEqual(Addressing(Addressing.INDIRECT, address=Integer(True, 4096)), results)

    def test_addressing_indexed(self):
        code = 'STA $1000,Y'
        results = self.parser.parse(code)[0].addressing
        self.assertEqual(Addressing(Addressing.INDEXED, address=Integer(True, 4096), register='Y'), results)

        code = 'LDA $C0,X'
        results = self.parser.parse(code)[0].addressing
        self.assertEqual(Addressing(Addressing.INDEXED, address=Integer(False, 192), register='X'), results)

    def test_addressing_indexed_indirect(self):
        code = 'LDA ($20,X)'
        results = self.parser.parse(code)[0].addressing
        self.assertEqual(Addressing(Addressing.INDEXED_INDIRECT, address=Integer(False, 32), register='X'), results)

    def test_addressing_indirect_indexed(self):
        code = 'LDA ($86),Y'
        results = self.parser.parse(code)[0].addressing
        self.assertEqual(Addressing(Addressing.INDIRECT_INDEXED, address=Integer(False, 134), register='Y'), results)



from typing import Union, List, Iterable
from functools import wraps

from .grammar import get_parser, Integer, Addressing, Arithmetic, Instruction


__all__ = ['Assembler', 'AssembleError']


class AssembleError(Exception):

    def __init__(self, info, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.info = info

    def __str__(self):
        return f'AssembleError: {self.info}'

    def __repr__(self):
        return f'AssembleError("{self.info}")'


CODE_MAP_IMPLIED = {
    'CLC': 0x18, 'CLD': 0xD8, 'CLI': 0x58, 'CLV': 0xB8, 'DEX': 0xCA,
    'DEY': 0x88, 'INX': 0xE8, 'INY': 0xC8, 'PHA': 0x48, 'PHP': 0x08,
    'PLA': 0x68, 'PLP': 0x28, 'RTI': 0x40, 'RTS': 0x60, 'SEC': 0x38,
    'SED': 0xF8, 'SEI': 0x78, 'TAX': 0xAA, 'TAY': 0xA8, 'TSX': 0xBA,
    'TXA': 0x8A, 'TXS': 0x9A, 'TYA': 0x98, 'JAM': 0x02,
}

CODE_MAP_RELATIVE = {
    'BCC': 0x90, 'BCS': 0xB0, 'BEQ': 0xF0, 'BMI': 0x30, 'BNE': 0xD0,
    'BPL': 0x10, 'BVC': 0x50, 'BVS': 0x70,
}

CODE_MAP_IMMEDIATE = {
    'ANC': 0x0B, 'ARR': 0x6B, 'ASR': 0x4B, 'SBX': 0xCB, 'XAA': 0x8B,
}

CODE_MAP_ABSOLUTE_Y = {
    'LAS': 0xBB, 'SHS': 0x9B, 'SHX': 0x9E,
}

CODE_MAPS_LOAD_A = {
    'ADC': {
        Addressing.IMMEDIATE: 0x69,
        Addressing.ZERO_PAGE: 0x65,
        Addressing.ZERO_PAGE_X: 0x75,
        Addressing.ABSOLUTE: 0x6D,
        Addressing.ABSOLUTE_X: 0x7D,
        Addressing.ABSOLUTE_Y: 0x79,
        Addressing.INDEXED_INDIRECT: 0x61,
        Addressing.INDIRECT_INDEXED: 0x71,
    },
    'AND': {
        Addressing.IMMEDIATE: 0x29,
        Addressing.ZERO_PAGE: 0x25,
        Addressing.ZERO_PAGE_X: 0x35,
        Addressing.ABSOLUTE: 0x2D,
        Addressing.ABSOLUTE_X: 0x3D,
        Addressing.ABSOLUTE_Y: 0x39,
        Addressing.INDEXED_INDIRECT: 0x21,
        Addressing.INDIRECT_INDEXED: 0x31,
    },
    'CMP': {
        Addressing.IMMEDIATE: 0xC9,
        Addressing.ZERO_PAGE: 0xC5,
        Addressing.ZERO_PAGE_X: 0xD5,
        Addressing.ABSOLUTE: 0xCD,
        Addressing.ABSOLUTE_X: 0xDD,
        Addressing.ABSOLUTE_Y: 0xD9,
        Addressing.INDEXED_INDIRECT: 0xC1,
        Addressing.INDIRECT_INDEXED: 0xD1,
    },
    'EOR': {
        Addressing.IMMEDIATE: 0x49,
        Addressing.ZERO_PAGE: 0x45,
        Addressing.ZERO_PAGE_X: 0x55,
        Addressing.ABSOLUTE: 0x4D,
        Addressing.ABSOLUTE_X: 0x5D,
        Addressing.ABSOLUTE_Y: 0x59,
        Addressing.INDEXED_INDIRECT: 0x41,
        Addressing.INDIRECT_INDEXED: 0x51,
    },
    'LDA': {
        Addressing.IMMEDIATE: 0xA9,
        Addressing.ZERO_PAGE: 0xA5,
        Addressing.ZERO_PAGE_X: 0xB5,
        Addressing.ABSOLUTE: 0xAD,
        Addressing.ABSOLUTE_X: 0xBD,
        Addressing.ABSOLUTE_Y: 0xB9,
        Addressing.INDEXED_INDIRECT: 0xA1,
        Addressing.INDIRECT_INDEXED: 0xB1,
    },
    'ORA': {
        Addressing.IMMEDIATE: 0x09,
        Addressing.ZERO_PAGE: 0x05,
        Addressing.ZERO_PAGE_X: 0x15,
        Addressing.ABSOLUTE: 0x0D,
        Addressing.ABSOLUTE_X: 0x1D,
        Addressing.ABSOLUTE_Y: 0x19,
        Addressing.INDEXED_INDIRECT: 0x01,
        Addressing.INDIRECT_INDEXED: 0x11,
    },
    'SBC': {
        Addressing.IMMEDIATE: 0xE9,
        Addressing.ZERO_PAGE: 0xE5,
        Addressing.ZERO_PAGE_X: 0xF5,
        Addressing.ABSOLUTE: 0xED,
        Addressing.ABSOLUTE_X: 0xFD,
        Addressing.ABSOLUTE_Y: 0xF9,
        Addressing.INDEXED_INDIRECT: 0xE1,
        Addressing.INDIRECT_INDEXED: 0xF1,
    },
}

CODE_MAPS_STORE_A = {
    'STA': {
        Addressing.ZERO_PAGE: 0x85,
        Addressing.ZERO_PAGE_X: 0x95,
        Addressing.ABSOLUTE: 0x8D,
        Addressing.ABSOLUTE_X: 0x9D,
        Addressing.ABSOLUTE_Y: 0x99,
        Addressing.INDEXED_INDIRECT: 0x81,
        Addressing.INDIRECT_INDEXED: 0x91,
    },
    'DCP': {
        Addressing.ZERO_PAGE: 0xC7,
        Addressing.ZERO_PAGE_X: 0xD7,
        Addressing.ABSOLUTE: 0xCF,
        Addressing.ABSOLUTE_X: 0xDF,
        Addressing.ABSOLUTE_Y: 0xDB,
        Addressing.INDEXED_INDIRECT: 0xC3,
        Addressing.INDIRECT_INDEXED: 0xD3,
    },
    'ISC': {
        Addressing.ZERO_PAGE: 0xE7,
        Addressing.ZERO_PAGE_X: 0xF7,
        Addressing.ABSOLUTE: 0xEF,
        Addressing.ABSOLUTE_X: 0xFF,
        Addressing.ABSOLUTE_Y: 0xFB,
        Addressing.INDEXED_INDIRECT: 0xE3,
        Addressing.INDIRECT_INDEXED: 0xF3,
    },
    'RLA': {
        Addressing.ZERO_PAGE: 0x27,
        Addressing.ZERO_PAGE_X: 0x37,
        Addressing.ABSOLUTE: 0x2F,
        Addressing.ABSOLUTE_X: 0x3F,
        Addressing.ABSOLUTE_Y: 0x3B,
        Addressing.INDEXED_INDIRECT: 0x23,
        Addressing.INDIRECT_INDEXED: 0x33,
    },
    'RRA': {
        Addressing.ZERO_PAGE: 0x67,
        Addressing.ZERO_PAGE_X: 0x77,
        Addressing.ABSOLUTE: 0x6F,
        Addressing.ABSOLUTE_X: 0x7F,
        Addressing.ABSOLUTE_Y: 0x7B,
        Addressing.INDEXED_INDIRECT: 0x63,
        Addressing.INDIRECT_INDEXED: 0x73,
    },
    'SLO': {
        Addressing.ZERO_PAGE: 0x07,
        Addressing.ZERO_PAGE_X: 0x17,
        Addressing.ABSOLUTE: 0x0F,
        Addressing.ABSOLUTE_X: 0x1F,
        Addressing.ABSOLUTE_Y: 0x1B,
        Addressing.INDEXED_INDIRECT: 0x03,
        Addressing.INDIRECT_INDEXED: 0x13,
    },
    'SRE': {
        Addressing.ZERO_PAGE: 0x47,
        Addressing.ZERO_PAGE_X: 0x57,
        Addressing.ABSOLUTE: 0x4F,
        Addressing.ABSOLUTE_X: 0x5F,
        Addressing.ABSOLUTE_Y: 0x5B,
        Addressing.INDEXED_INDIRECT: 0x43,
        Addressing.INDIRECT_INDEXED: 0x53,
    },
}


CODE_MAPS_A_M = {
    'ASL': {
        Addressing.ACCUMULATOR: 0x0A,
        Addressing.ZERO_PAGE: 0x06,
        Addressing.ZERO_PAGE_X: 0x16,
        Addressing.ABSOLUTE: 0x0E,
        Addressing.ABSOLUTE_X: 0x1E,
    },
    'LSR': {
        Addressing.ACCUMULATOR: 0x4A,
        Addressing.ZERO_PAGE: 0x46,
        Addressing.ZERO_PAGE_X: 0x56,
        Addressing.ABSOLUTE: 0x4E,
        Addressing.ABSOLUTE_X: 0x5E,
    },
    'ROL': {
        Addressing.ACCUMULATOR: 0x2A,
        Addressing.ZERO_PAGE: 0x26,
        Addressing.ZERO_PAGE_X: 0x36,
        Addressing.ABSOLUTE: 0x2E,
        Addressing.ABSOLUTE_X: 0x3E,
    },
    'ROR': {
        Addressing.ACCUMULATOR: 0x6A,
        Addressing.ZERO_PAGE: 0x66,
        Addressing.ZERO_PAGE_X: 0x76,
        Addressing.ABSOLUTE: 0x6E,
        Addressing.ABSOLUTE_X: 0x7E,
    },
}


class Assembler(object):

    def __init__(self,
                 max_memory=0x10000,
                 program_entry=0xfffc,
                 brk_size=2):
        self.max_memory = max_memory
        self.program_entry = program_entry
        self.brk_size = brk_size
        assert brk_size in {1, 2}, 'The size of BRK should be in {1, 2}'

        self.code_start = -1  # The offset of the first instruction that can be executed
        self.code_offset = 0  # Current offset
        self.line_number = -1  # Current line number
        self.code_offsets = []  # The offsets of all the instructions
        self.fit_zero_pages = []  # Whether the addresses fit zero-page
        self.label_offsets = {}  # The resolved labels
        self.codes = []  # The generated codes

    def reset(self):
        self.code_start = -1
        self.code_offset = 0
        self.line_number = -1
        self.code_offsets = []
        self.fit_zero_pages = []
        self.label_offsets = {}
        self.codes = []

    def assemble(self,
                 instructions: Union[str, List],
                 add_entry: bool = True):
        if isinstance(instructions, str):
            parser = get_parser()
            instructions = parser.parse(instructions)
        # Preprocess and calculate offsets
        self.reset()
        for i, inst in enumerate(instructions):
            self.line_number = inst.line_num
            if inst.label is not None and not inst.op.endswith('ORG'):
                self.label_offsets[inst.label] = self.code_offset
            op_name = inst.op.lower()
            if op_name.startswith('.'):
                op_name = op_name[1:]
            if inst.op in CODE_MAP_IMPLIED:
                offset = self._get_num_bytes_type_implied(inst.addressing, op_name)
            elif inst.op in CODE_MAP_IMMEDIATE:
                offset = self._get_num_bytes_type_immediate(inst.addressing, op_name)
            elif inst.op in CODE_MAP_RELATIVE:
                offset = self._get_num_bytes_type_relative(inst.addressing, op_name)
            elif inst.op in CODE_MAP_ABSOLUTE_Y:
                offset = self._get_num_bytes_type_absolute_y(inst.addressing, op_name)
            elif inst.op in CODE_MAPS_LOAD_A:
                offset = self._get_num_bytes_type_load_a(inst.addressing, op_name)
            elif inst.op in CODE_MAPS_STORE_A:
                offset = self._get_num_bytes_type_store_a(inst.addressing, op_name)
            elif inst.op in CODE_MAPS_A_M:
                offset = self._get_num_bytes_type_a_m(inst.addressing, op_name)
            else:
                offset = getattr(self, f'pre_{op_name}')(inst.addressing, op_name)
            if inst.label is not None and inst.op.endswith('ORG'):
                self.label_offsets[inst.label] = self.code_offset
            if self.code_start == -1 and inst.op in Instruction.KEYWORDS:
                self.code_start = self.code_offset
            self.code_offsets.append(self.code_offset)
            self.code_offset += offset
            if self.code_offset >= self.max_memory:
                raise AssembleError(f"The assembled code will exceed the "
                                    f"max memory {hex(self.max_memory)} "
                                    f"at line {self.line_number}")
        # Generate codes
        for i, inst in enumerate(instructions):
            self.line_number = inst.line_num
            self.code_offset = self.code_offsets[i]
            if inst.op in CODE_MAP_IMPLIED:
                self._extend_address_type_implied(i, inst.addressing, inst.op)
            elif inst.op in CODE_MAP_IMMEDIATE:
                self._extend_address_type_immediate(i, inst.addressing, inst.op)
            elif inst.op in CODE_MAP_RELATIVE:
                self._extend_address_type_relative(i, inst.addressing, inst.op)
            elif inst.op in CODE_MAP_ABSOLUTE_Y:
                self._extend_address_type_absolute_y(i, inst.addressing, inst.op)
            elif inst.op in CODE_MAPS_LOAD_A:
                self._extend_address_type_load_a(i, inst.addressing, inst.op)
            elif inst.op in CODE_MAPS_STORE_A:
                self._extend_address_type_store_a(i, inst.addressing, inst.op)
            elif inst.op in CODE_MAPS_A_M:
                self._extend_address_type_a_m(i, inst.addressing, inst.op)
            else:
                op_name = inst.op.lower()
                if op_name.startswith('.'):
                    op_name = op_name[1:]
                getattr(self, f'gen_{op_name}')(i, inst.addressing)
        while len(self.codes) and len(self.codes[-1][1]) == 0:
            del self.codes[-1]
        if add_entry:
            self.code_offset = 0xFFFC
            self.gen_entry(None, Addressing(Addressing.ADDRESS, address=Integer(is_word=True, value=self.code_start)))
        return self.codes

    def _addressing_guard(allowed: Iterable[str]):
        def deco(func):
            @wraps(func)
            def inner(self, addressing, op_name):
                if addressing.mode not in allowed:
                    raise AssembleError(f"{addressing.mode.capitalize()} addressing is not allowed"
                                        f" for `{op_name.upper()}` at line {self.line_number}")
                self.fit_zero_pages.append(False)
                if addressing.mode in {Addressing.ADDRESS, Addressing.INDEXED}:
                    try:
                        resolved = self._resolve_address(addressing)
                        if isinstance(resolved.address, Integer) and \
                           not resolved.address.is_word and resolved.address.value <= 0xFF:
                            self.fit_zero_pages[-1] = True
                    except AssembleError as e:
                        pass
                return func(self, addressing)
            return inner
        return deco

    def _assemble_guard(func):
        def inner(self, index, addressing, *args, **kwargs):
            while len(self.codes) and len(self.codes[-1][1]) == 0:
                del self.codes[-1]
            if len(self.codes) == 0 or self.code_offset != self.codes[-1][0] + len(self.codes[-1][1]):
                self.codes.append((self.code_offset, []))
            addressing = self._resolve_address(addressing)
            if addressing.mode in {Addressing.IMMEDIATE, Addressing.INDEXED_INDIRECT, Addressing.INDIRECT_INDEXED} \
                    and addressing.address.value > 0xFF:
                raise AssembleError(f"The value {hex(addressing.address.value)} is too large for the addressing "
                                    f"at line {self.line_number}")
            return func(self, index, addressing, *args, **kwargs)
        return inner

    def _resolve_address_recur(self, arithmetic: Union[Integer, Arithmetic]) -> Union[Integer, List[Integer]]:
        if arithmetic is None:
            return None
        if isinstance(arithmetic, Integer):
            return arithmetic
        if arithmetic.mode == Arithmetic.CURRENT:
            return Integer(is_word=True, value=self.code_offset)
        if arithmetic.mode == Arithmetic.LABEL:
            if arithmetic.param not in self.label_offsets:
                raise AssembleError(f"Can not resolve label '{arithmetic.param}' at line {self.line_number}")
            return Integer(is_word=True, value=self.label_offsets[arithmetic.param])
        if arithmetic.mode == Arithmetic.ADD:
            return self._resolve_address_recur(arithmetic.param[0]) + self._resolve_address_recur(arithmetic.param[1])
        if arithmetic.mode == Arithmetic.SUB:
            return self._resolve_address_recur(arithmetic.param[0]) - self._resolve_address_recur(arithmetic.param[1])
        if arithmetic.mode == Arithmetic.MUL:
            return self._resolve_address_recur(arithmetic.param[0]) * self._resolve_address_recur(arithmetic.param[1])
        if arithmetic.mode == Arithmetic.DIV:
            return self._resolve_address_recur(arithmetic.param[0]) // self._resolve_address_recur(arithmetic.param[1])
        if arithmetic.mode == Arithmetic.NEG:
            return -self._resolve_address_recur(arithmetic.param)
        if arithmetic.mode == Arithmetic.LOW_BYTE:
            return self._resolve_address_recur(arithmetic.param).low_byte()
        if arithmetic.mode == Arithmetic.HIGH_BYTE:
            return self._resolve_address_recur(arithmetic.param).high_byte()
        if arithmetic.mode == Arithmetic.LIST:
            return [self._resolve_address_recur(p) for p in arithmetic.param]

    def _resolve_address(self, addressing: Addressing) -> Addressing:
        return Addressing(mode=addressing.mode,
                          address=self._resolve_address_recur(addressing.address),
                          register=addressing.register)

    def _extend_byte(self, code):
        self.codes[-1][1].append(code)

    def _extend_byte_address(self, code, addressing: Addressing):
        self.codes[-1][1].extend([code, addressing.address.value])

    def _extend_word_address(self, code, addressing: Addressing):
        self.codes[-1][1].extend([code, addressing.address.low_byte().value, addressing.address.high_byte().value])

    @_addressing_guard(allowed={Addressing.IMPLIED})
    def _get_num_bytes_type_implied(self, addressing: Addressing):
        return 1

    @_assemble_guard
    def _extend_address_type_implied(self, index, addressing: Addressing, op: str):
        self._extend_byte(CODE_MAP_IMPLIED[op])

    @_addressing_guard(allowed={Addressing.IMMEDIATE})
    def _get_num_bytes_type_immediate(self, addressing: Addressing):
        return 2

    @_assemble_guard
    def _extend_address_type_immediate(self, index, addressing: Addressing, op: str):
        self._extend_byte_address(CODE_MAP_IMMEDIATE[op], addressing)

    @_addressing_guard(allowed={Addressing.ADDRESS})
    def _get_num_bytes_type_relative(self, addressing: Addressing):
        return 2

    @_assemble_guard
    def _extend_address_type_relative(self, index, addressing: Addressing, op: str):
        address = addressing.address.value - (self.codes[-1][0] + 2)
        if address < -0x80 or 0x7F < address:
            raise AssembleError(f"The offset {hex(address)} is out of range for relative addressing "
                                f"at line {self.line_number}")
        if address < 0:
            address += 0x100
        self.codes[-1][1].extend([CODE_MAP_RELATIVE[op], address])

    @_addressing_guard(allowed={Addressing.INDEXED})
    def _get_num_bytes_type_absolute_y(self, addressing: Addressing):
        return 3

    @_assemble_guard
    def _extend_address_type_absolute_y(self, index, addressing: Addressing, op: str):
        self._extend_word_address(CODE_MAP_ABSOLUTE_Y[op], addressing)

    @_addressing_guard(allowed={Addressing.IMMEDIATE, Addressing.ADDRESS, Addressing.INDEXED,
                                Addressing.INDEXED_INDIRECT, Addressing.INDIRECT_INDEXED})
    def _get_num_bytes_type_load_a(self, addressing: Addressing):
        if addressing.mode in {Addressing.ADDRESS, Addressing.INDEXED}:
            return 2 if self.fit_zero_pages[-1] else 3
        return 2

    @_assemble_guard
    def _extend_address_type_load_a(self, index, addressing: Addressing, op: str):
        code_map = CODE_MAPS_LOAD_A[op]
        if addressing.mode == Addressing.IMMEDIATE:
            self._extend_byte_address(code_map[Addressing.IMMEDIATE], addressing)
        elif addressing.mode == Addressing.ADDRESS:
            if self.fit_zero_pages[index]:
                self._extend_byte_address(code_map[Addressing.ZERO_PAGE], addressing)
            else:
                self._extend_word_address(code_map[Addressing.ABSOLUTE], addressing)
        elif addressing.mode == Addressing.INDEXED:
            if self.fit_zero_pages[index] and addressing.register == 'X':
                self._extend_byte_address(code_map[Addressing.ZERO_PAGE_X], addressing)
            elif addressing.register == 'X':
                self._extend_word_address(code_map[Addressing.ABSOLUTE_X], addressing)
            else:
                self._extend_word_address(code_map[Addressing.ABSOLUTE_Y], addressing)
        elif addressing.mode == Addressing.INDEXED_INDIRECT:
            self._extend_byte_address(code_map[Addressing.INDEXED_INDIRECT], addressing)
        elif addressing.mode == Addressing.INDIRECT_INDEXED:
            self._extend_byte_address(code_map[Addressing.INDIRECT_INDEXED], addressing)

    @_addressing_guard(allowed={Addressing.ADDRESS, Addressing.INDEXED,
                                Addressing.INDEXED_INDIRECT, Addressing.INDIRECT_INDEXED})
    def _get_num_bytes_type_store_a(self, addressing: Addressing):
        if addressing.mode in {Addressing.ADDRESS, Addressing.INDEXED}:
            return 2 if self.fit_zero_pages[-1] else 3
        return 2

    @_assemble_guard
    def _extend_address_type_store_a(self, index, addressing: Addressing, op: str):
        code_map = CODE_MAPS_STORE_A[op]
        if addressing.mode == Addressing.ADDRESS:
            if self.fit_zero_pages[index]:
                self._extend_byte_address(code_map[Addressing.ZERO_PAGE], addressing)
            else:
                self._extend_word_address(code_map[Addressing.ABSOLUTE], addressing)
        elif addressing.mode == Addressing.INDEXED:
            if self.fit_zero_pages[index] and addressing.register == 'X':
                self._extend_byte_address(code_map[Addressing.ZERO_PAGE_X], addressing)
            elif addressing.register == 'X':
                self._extend_word_address(code_map[Addressing.ABSOLUTE_X], addressing)
            else:
                self._extend_word_address(code_map[Addressing.ABSOLUTE_Y], addressing)
        elif addressing.mode == Addressing.INDEXED_INDIRECT:
            self._extend_byte_address(code_map[Addressing.INDEXED_INDIRECT], addressing)
        elif addressing.mode == Addressing.INDIRECT_INDEXED:
            self._extend_byte_address(code_map[Addressing.INDIRECT_INDEXED], addressing)

    @_addressing_guard(allowed={Addressing.ACCUMULATOR, Addressing.ADDRESS, Addressing.INDEXED})
    def _get_num_bytes_type_a_m(self, addressing: Addressing):
        if addressing.mode == Addressing.ACCUMULATOR:
            return 1
        return 2 if self.fit_zero_pages[-1] else 3

    @_assemble_guard
    def _extend_address_type_a_m(self, index, addressing: Addressing, op: str):
        code_map = CODE_MAPS_A_M[op]
        if addressing.mode == Addressing.ACCUMULATOR:
            self._extend_byte(code_map[Addressing.ACCUMULATOR])
        elif addressing.mode == Addressing.ADDRESS:
            if self.fit_zero_pages[index]:
                self._extend_byte_address(code_map[Addressing.ZERO_PAGE], addressing)
            else:
                self._extend_word_address(code_map[Addressing.ABSOLUTE], addressing)
        elif addressing.mode == Addressing.INDEXED:
            if addressing.register == 'Y':
                raise AssembleError(f"Can not use Y as the index register in {op} at line {self.line_number}")
            if self.fit_zero_pages[index]:
                self._extend_byte_address(code_map[Addressing.ZERO_PAGE_X], addressing)
            else:
                self._extend_word_address(code_map[Addressing.ABSOLUTE_X], addressing)

    @_assemble_guard
    def gen_entry(self, index, addressing: Addressing):
        self.codes[-1][1].extend([addressing.address.low_byte().value, addressing.address.high_byte().value])

    @_addressing_guard(allowed={Addressing.ADDRESS})
    def pre_org(self, addressing: Addressing):
        self.code_offset = self._resolve_address(addressing).address.value
        return 0

    @_assemble_guard
    def gen_org(self, index, addressing: Addressing):
        pass

    @_addressing_guard(allowed={Addressing.IMPLIED})
    def pre_end(self, addressing: Addressing):
        return 3

    @_assemble_guard
    def gen_end(self, index, addressing: Addressing):
        address = Integer(is_word=True, value=self.codes[-1][0])
        self.codes[-1][1].extend([0x4C, address.low_byte().value, address.high_byte().value])

    @_addressing_guard(allowed={Addressing.ADDRESS, Addressing.LIST})
    def pre_byte(self, addressing: Addressing):
        if addressing.mode == Addressing.LIST:
            return len(addressing.address.param)
        return 1

    @_assemble_guard
    def gen_byte(self, index, addressing: Addressing):
        if addressing.mode == Addressing.LIST:
            for byte in addressing.address:
                if byte.value > 0xFF:
                    raise AssembleError(f"{hex(byte.value)} can not fit in a byte "
                                        f"at line {self.line_number}")
                self._extend_byte(byte.value)
        else:
            if addressing.address.value > 0xFF:
                raise AssembleError(f"{hex(addressing.address.value)} can not fit in a byte "
                                    f"at line {self.line_number}")
            self._extend_byte(addressing.address.value)

    @_addressing_guard(allowed={Addressing.ADDRESS, Addressing.LIST})
    def pre_word(self, addressing: Addressing):
        if addressing.mode == Addressing.LIST:
            return len(addressing.address.param) * 2
        return 2

    @_assemble_guard
    def gen_word(self, index, addressing: Addressing):
        if addressing.mode == Addressing.LIST:
            for word in addressing.address:
                self.codes[-1][1].extend([word.low_byte().value, word.high_byte().value])
        else:
            self.codes[-1][1].extend([addressing.address.low_byte().value, addressing.address.high_byte().value])

    @_addressing_guard(allowed={Addressing.IMPLIED})
    def pre_brk(self, addressing: Addressing):
        return self.brk_size

    @_assemble_guard
    def gen_brk(self, index, addressing: Addressing):
        for i in range(self.brk_size):
            self._extend_byte(0x00)

    @_addressing_guard(allowed={Addressing.IMPLIED, Addressing.IMMEDIATE, Addressing.ADDRESS, Addressing.INDEXED})
    def pre_nop(self, addressing: Addressing):
        if addressing.mode == Addressing.IMPLIED:
            return 1
        if addressing.mode in {Addressing.ADDRESS, Addressing.INDEXED}:
            return 2 if self.fit_zero_pages[-1] else 3
        return 2

    @_assemble_guard
    def gen_nop(self, index, addressing: Addressing):
        if addressing.mode == Addressing.IMPLIED:
            self._extend_byte(0xEA)
        elif addressing.mode == Addressing.IMMEDIATE:
            self._extend_byte_address(0x80, addressing)
        elif addressing.mode == Addressing.ADDRESS:
            if self.fit_zero_pages[index]:
                self._extend_byte_address(0x04, addressing)
            else:
                self._extend_word_address(0x0C, addressing)
        elif addressing.mode == Addressing.INDEXED:
            if addressing.register == 'Y':
                raise AssembleError(f"Can not use Y as the index register in NOP at line {self.line_number}")
            if self.fit_zero_pages[index]:
                self._extend_byte_address(0x14, addressing)
            else:
                self._extend_word_address(0x1C, addressing)

    @_addressing_guard(allowed={Addressing.ADDRESS, Addressing.INDIRECT})
    def pre_jmp(self, addressing: Addressing):
        return 3

    @_assemble_guard
    def gen_jmp(self, index, addressing: Addressing):
        if addressing.mode == Addressing.ADDRESS:
            self._extend_word_address(0x4C, addressing)
        elif addressing.mode == Addressing.INDIRECT:
            self._extend_word_address(0x6C, addressing)

    @_addressing_guard(allowed={Addressing.ADDRESS})
    def pre_jsr(self, addressing: Addressing):
        return 3

    @_assemble_guard
    def gen_jsr(self, index, addressing: Addressing):
        self._extend_word_address(0x20, addressing)

    @_addressing_guard(allowed={Addressing.IMMEDIATE, Addressing.ADDRESS, Addressing.INDEXED})
    def pre_ldx(self, addressing: Addressing):
        if addressing.mode in {Addressing.ADDRESS, Addressing.INDEXED}:
            return 2 if self.fit_zero_pages[-1] else 3
        return 2

    @_assemble_guard
    def gen_ldx(self, index, addressing: Addressing):
        if addressing.mode == Addressing.IMMEDIATE:
            self._extend_byte_address(0xA2, addressing)
        elif addressing.mode == Addressing.ADDRESS:
            if self.fit_zero_pages[index]:
                self._extend_byte_address(0xA6, addressing)
            else:
                self._extend_word_address(0xAE, addressing)
        elif addressing.mode == Addressing.INDEXED:
            if addressing.register == 'X':
                raise AssembleError(f"Can not use X as the index register in LDX at line {self.line_number}")
            if self.fit_zero_pages[index]:
                self._extend_byte_address(0xB6, addressing)
            else:
                self._extend_word_address(0xBE, addressing)

    @_addressing_guard(allowed={Addressing.IMMEDIATE, Addressing.ADDRESS, Addressing.INDEXED})
    def pre_ldy(self, addressing: Addressing):
        if addressing.mode in {Addressing.ADDRESS, Addressing.INDEXED}:
            return 2 if self.fit_zero_pages[-1] else 3
        return 2

    @_assemble_guard
    def gen_ldy(self, index, addressing: Addressing):
        if addressing.mode == Addressing.IMMEDIATE:
            self._extend_byte_address(0xA0, addressing)
        elif addressing.mode == Addressing.ADDRESS:
            if self.fit_zero_pages[index]:
                self._extend_byte_address(0xA4, addressing)
            else:
                self._extend_word_address(0xAC, addressing)
        elif addressing.mode == Addressing.INDEXED:
            if addressing.register == 'Y':
                raise AssembleError(f"Can not use Y as the index register in LDY at line {self.line_number}")
            if self.fit_zero_pages[index]:
                self._extend_byte_address(0xB4, addressing)
            else:
                self._extend_word_address(0xBC, addressing)

    @_addressing_guard(allowed={Addressing.IMMEDIATE, Addressing.ADDRESS, Addressing.INDEXED,
                                Addressing.INDIRECT_INDEXED, Addressing.INDEXED_INDIRECT})
    def pre_lax(self, addressing: Addressing):
        if addressing.mode in {Addressing.ADDRESS, Addressing.INDEXED}:
            return 2 if self.fit_zero_pages[-1] else 3
        return 2

    @_assemble_guard
    def gen_lax(self, index, addressing: Addressing):
        if addressing.mode == Addressing.IMMEDIATE:
            self._extend_byte_address(0xAB, addressing)
        elif addressing.mode == Addressing.ADDRESS:
            if self.fit_zero_pages[index]:
                self._extend_byte_address(0xA7, addressing)
            else:
                self._extend_word_address(0xAF, addressing)
        elif addressing.mode == Addressing.INDEXED:
            if addressing.register == 'X':
                raise AssembleError(f"Can not use X as the index register in LAX at line {self.line_number}")
            if self.fit_zero_pages[index]:
                self._extend_byte_address(0xB7, addressing)
            else:
                self._extend_word_address(0xBF, addressing)
        elif addressing.mode == Addressing.INDEXED_INDIRECT:
            self._extend_byte_address(0xA3, addressing)
        elif addressing.mode == Addressing.INDIRECT_INDEXED:
            self._extend_byte_address(0xB3, addressing)

    @_addressing_guard(allowed={Addressing.ADDRESS, Addressing.INDEXED})
    def pre_stx(self, addressing: Addressing):
        return 2 if self.fit_zero_pages[-1] else 3

    @_assemble_guard
    def gen_stx(self, index, addressing: Addressing):
        if addressing.mode == Addressing.ADDRESS:
            if self.fit_zero_pages[index]:
                self._extend_byte_address(0x86, addressing)
            else:
                self._extend_word_address(0x8E, addressing)
        elif addressing.mode == Addressing.INDEXED:
            if addressing.register == 'X':
                raise AssembleError(f"Can not use X as the index register in STX at line {self.line_number}")
            if addressing.address.value > 0xFF:
                raise AssembleError(f"Absolute indexed addressing is not allowed for STX "
                                    f"at line {self.line_number}")
            self._extend_byte_address(0x96, addressing)

    @_addressing_guard(allowed={Addressing.ADDRESS, Addressing.INDEXED})
    def pre_sty(self, addressing: Addressing):
        return 2 if self.fit_zero_pages[-1] else 3

    @_assemble_guard
    def gen_sty(self, index, addressing: Addressing):
        if addressing.mode == Addressing.ADDRESS:
            if self.fit_zero_pages[index]:
                self._extend_byte_address(0x84, addressing)
            else:
                self._extend_word_address(0x8C, addressing)
        elif addressing.mode == Addressing.INDEXED:
            if addressing.register == 'Y':
                raise AssembleError(f"Can not use Y as the index register in STY at line {self.line_number}")
            if addressing.address.value > 0xFF:
                raise AssembleError(f"Absolute indexed addressing is not allowed for STY "
                                    f"at line {self.line_number}")
            self._extend_byte_address(0x94, addressing)

    @_addressing_guard(allowed={Addressing.ADDRESS, Addressing.INDEXED, Addressing.INDEXED_INDIRECT})
    def pre_sax(self, addressing: Addressing):
        if addressing.mode in {Addressing.ADDRESS, Addressing.INDEXED}:
            return 2 if self.fit_zero_pages[-1] else 3
        return 2

    @_assemble_guard
    def gen_sax(self, index, addressing: Addressing):
        if addressing.mode == Addressing.ADDRESS:
            if self.fit_zero_pages[index]:
                self._extend_byte_address(0x87, addressing)
            else:
                self._extend_word_address(0x8F, addressing)
        elif addressing.mode == Addressing.INDEXED:
            if addressing.register == 'X':
                raise AssembleError(f"Can not use X as the index register in SAX at line {self.line_number}")
            if addressing.address.value > 0xFF:
                raise AssembleError(f"Absolute indexed addressing is not allowed for SAX "
                                    f"at line {self.line_number}")
            self._extend_byte_address(0x97, addressing)
        elif addressing.mode == Addressing.INDEXED_INDIRECT:
            self._extend_byte_address(0x83, addressing)

    @_addressing_guard(allowed={Addressing.INDEXED, Addressing.INDIRECT_INDEXED})
    def pre_sha(self, addressing: Addressing):
        return 3 if addressing.mode == Addressing.INDEXED else 2

    @_assemble_guard
    def gen_sha(self, index, addressing: Addressing):
        if addressing.mode == Addressing.INDEXED:
            if addressing.register == 'X':
                raise AssembleError(f"Can not use X as the index register in SHA at line {self.line_number}")
            self._extend_word_address(0x9F, addressing)
        elif addressing.mode == Addressing.INDIRECT_INDEXED:
            self._extend_byte_address(0x93, addressing)

    @_addressing_guard(allowed={Addressing.INDEXED})
    def pre_shy(self, addressing: Addressing):
        return 3 if addressing.mode == Addressing.INDEXED else 2

    @_assemble_guard
    def gen_shy(self, index, addressing: Addressing):
        if addressing.mode == Addressing.INDEXED:
            if addressing.register == 'Y':
                raise AssembleError(f"Can not use Y as the index register in SHY at line {self.line_number}")
            self._extend_word_address(0x9C, addressing)

    @_addressing_guard(allowed={Addressing.ADDRESS})
    def pre_bit(self, addressing: Addressing):
        return 2 if self.fit_zero_pages[-1] else 3

    @_assemble_guard
    def gen_bit(self, index, addressing: Addressing):
        if self.fit_zero_pages[index]:
            self._extend_byte_address(0x24, addressing)
        else:
            self._extend_word_address(0x2C, addressing)

    @_addressing_guard(allowed={Addressing.IMMEDIATE, Addressing.ADDRESS})
    def pre_cpx(self, addressing: Addressing):
        if addressing.mode in {Addressing.ADDRESS}:
            return 2 if self.fit_zero_pages[-1] else 3
        return 2

    @_assemble_guard
    def gen_cpx(self, index, addressing: Addressing):
        if addressing.mode == Addressing.IMMEDIATE:
            self._extend_byte_address(0xE0, addressing)
        elif addressing.mode == Addressing.ADDRESS:
            if self.fit_zero_pages[index]:
                self._extend_byte_address(0xE4, addressing)
            else:
                self._extend_word_address(0xEC, addressing)

    @_addressing_guard(allowed={Addressing.IMMEDIATE, Addressing.ADDRESS})
    def pre_cpy(self, addressing: Addressing):
        if addressing.mode in {Addressing.ADDRESS}:
            return 2 if self.fit_zero_pages[-1] else 3
        return 2

    @_assemble_guard
    def gen_cpy(self, index, addressing: Addressing):
        if addressing.mode == Addressing.IMMEDIATE:
            self._extend_byte_address(0xC0, addressing)
        elif addressing.mode == Addressing.ADDRESS:
            if self.fit_zero_pages[index]:
                self._extend_byte_address(0xC4, addressing)
            else:
                self._extend_word_address(0xCC, addressing)

    @_addressing_guard(allowed={Addressing.ADDRESS, Addressing.INDEXED})
    def pre_inc(self, addressing: Addressing):
        return 2 if self.fit_zero_pages[-1] else 3

    @_assemble_guard
    def gen_inc(self, index, addressing: Addressing):
        if addressing.mode == Addressing.ADDRESS:
            if self.fit_zero_pages[index]:
                self._extend_byte_address(0xE6, addressing)
            else:
                self._extend_word_address(0xEE, addressing)
        elif addressing.mode == Addressing.INDEXED:
            if addressing.register == 'Y':
                raise AssembleError(f"Can not use Y as the index register in INC at line {self.line_number}")
            if self.fit_zero_pages[index]:
                self._extend_byte_address(0xF6, addressing)
            else:
                self._extend_word_address(0xFE, addressing)

    @_addressing_guard(allowed={Addressing.ADDRESS, Addressing.INDEXED})
    def pre_dec(self, addressing: Addressing):
        return 2 if self.fit_zero_pages[-1] else 3

    @_assemble_guard
    def gen_dec(self, index, addressing: Addressing):
        if addressing.mode == Addressing.ADDRESS:
            if self.fit_zero_pages[index]:
                self._extend_byte_address(0xC6, addressing)
            else:
                self._extend_word_address(0xCE, addressing)
        elif addressing.mode == Addressing.INDEXED:
            if addressing.register == 'Y':
                raise AssembleError(f"Can not use Y as the index register in DEC at line {self.line_number}")
            if self.fit_zero_pages[index]:
                self._extend_byte_address(0xD6, addressing)
            else:
                self._extend_word_address(0xDE, addressing)



from unittest import TestCase

from asm_6502 import get_parser, ParseError


class TestParseError(TestCase):

    def setUp(self) -> None:
        self.parser = get_parser()

    def test_illegal_character(self):
        code = 'ORG  $@0080'
        with self.assertRaises(ParseError) as e:
            self.parser.parse(code)
        self.assertEqual("ParseError: Illegal character '$' found at line 1, column 6", str(e.exception))

        code = 'ORG  $0080\nORG $@0800'
        with self.assertRaises(ParseError) as e:
            self.parser.parse(code)
        self.assertEqual("ParseError(\"Illegal character '$' found at line 2, column 5\")", repr(e.exception))

    def test_parse_comma(self):
        code = 'LDA $0080,'
        with self.assertRaises(ParseError) as e:
            self.parser.parse(code)
        self.assertEqual("ParseError: Syntax error at EOF", str(e.exception))

    def test_parse_too_many_parameters(self):
        code = 'LDA XXX,YYY,ZZZ ZZZ'
        with self.assertRaises(ParseError) as e:
            self.parser.parse(code)
        self.assertEqual("ParseError: Syntax error at line 1, column 17: 'ZZZ'", str(e.exception))

    def test_wrong_accumulator(self):
        code = 'LSR X'
        with self.assertRaises(ParseError) as e:
            self.parser.parse(code)
        self.assertEqual("ParseError: Register X can not be used as an address at 1, column 5", str(e.exception))

    def test_wrong_register_in_indexed_addressing(self):
        code = 'STA $1000,A'
        with self.assertRaises(ParseError) as e:
            self.parser.parse(code)
        self.assertEqual("ParseError: Only registers X and Y can be used for indexing, found A at 1, column 11",
                         str(e.exception))

    def test_wrong_register_in_indexed_indirect_addressing(self):
        code = 'LDA ($20,Y)'
        with self.assertRaises(ParseError) as e:
            self.parser.parse(code)
        self.assertEqual("ParseError: Only register X can be used for indexed indirect addressing, "
                         "found Y at 1, column 10",
                         str(e.exception))

    def test_wrong_register_in_indirect_indexed_addressing(self):
        code = 'LDA ($86),X'
        with self.assertRaises(ParseError) as e:
            self.parser.parse(code)
        self.assertEqual("ParseError: Only register Y can be used for indexed indirect addressing, "
                         "found X at 1, column 11",
                         str(e.exception))
